from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map, str
from future.moves.urllib import request
from future.moves.urllib.error import URLError

import os, os.path, logging, gzip, errno, json, socket
from io import open
from typing import TYPE_CHECKING

from .compat import FileExistsError
from .file import copyfilelike, Tell
from .iter import first_not_none
from .filesystem import safe_filename
from .exceptions import DownloadFailed
from .url import get_filename_from_url

if TYPE_CHECKING:
	from typing import Callable, Optional, Tuple, Mapping
	from http.cookiejar import CookieJar
	from http.client import HTTPMessage

if __debug__:
	import requests

logger = logging.getLogger(__name__)

class HTTPError(Exception):

	def __init__(self, *args, **kwargs):
		response = kwargs.pop("response", None)
		Exception.__init__(self, *args, **kwargs)
		self.response = response

class ContentInvalidLength(HTTPError, DownloadFailed):

	def __init__(self, path, expected, received):
		HTTPError.__init__(self, path, expected, received)

class TimeOut(HTTPError, DownloadFailed):
	pass

class DownloadInterrupted(HTTPError, DownloadFailed):
	pass

class NoRedirect(HTTPError):
	pass

try:
	from email.utils import parsedate_to_datetime # new in 3.3

	def parsedate_to_timestamp(datestr): # correct timezone
		return parsedate_to_datetime(datestr).timestamp()

except ImportError:
	import time

	def parsedate_to_timestamp(datestr): # fixme: incorrect timezone
		# type: (str, ) -> float

		try:
			#str fixed py2.x unicode bug
			return time.mktime(time.strptime(datestr, str("%a, %d %b %Y %H:%M:%S %Z")))
		except ValueError:
			#time data 'Thu, 02 Aug 2007 06:54:23 GMT' does not match format u'%a, %d %b %Y %H:%M:%S +0000' # fixed with str()
			return time.mktime(time.strptime(datestr, str("%a, %d %b %Y %H:%M:%S +0000")))

try:
	from email.message import EmailMessage # new in 3.6

	def get_filename(headers):
		# type: (EmailMessage, ) -> str

		return headers.get_filename()

except ImportError:
	def get_filename(headers):
		# type: (Mapping, ) -> str

		try:
			content_disposition = headers["Content-Disposition"]
			return content_disposition.split("=")[1].replace("\"", "").strip()
		except (KeyError, IndexError):
			return None

class URLRequestBuilder(object):

	def __init__(self, cookiejar=None, basicauth=None):
		# type: (Optional[CookieJar], Optional[Tuple[str, str, str]]) -> None

		handlers = []

		if cookiejar:
			handlers.append(request.HTTPCookieProcessor(cookiejar))
		if basicauth:
			top_level_url, username, password = basicauth
			password_mgr = request.HTTPPasswordMgrWithDefaultRealm()
			password_mgr.add_password(None, top_level_url, username, password)
			handlers.append(request.HTTPBasicAuthHandler(password_mgr))

		if handlers:
			opener = request.build_opener(handlers)
			self.openfunc = opener.open
		else:
			self.openfunc = request.urlopen

	def getfunc(self):
		return self.openfunc

	def request(self, url, headers=None, timeout=None):
		return URLRequest(url, headers, timeout, openfunc=self.openfunc)

class FileLike(object):

	def __init__(self, urlrequest):
		self.urlrequest = urlrequest

		self.content_length = self.urlrequest._content_length()

		self.response = Tell(self.urlrequest.response)
		if self.urlrequest.headers.get("Content-Encoding", None) == "gzip":
			# if that does not work, use BytesIO
			logger.info("Using gzip")
			self.fp = gzip.GzipFile(fileobj=self.response)
		else:
			self.fp = self.response

	def close(self):
		self.fp.close()

		transferred = self.response.tell() # is this correct?
		if self.content_length is not None and self.content_length != transferred:
			raise ContentInvalidLength(None, self.content_length, transferred)

	def __enter__(self):
		return self.fp

	def __exit__(self, exc_type, exc_value, traceback):
		self.close()

def get_redirect_url(url, headers=None):

	import requests

	r = requests.get(url, allow_redirects=False, headers=headers)
	r.raise_for_status()

	try:
		if r.status_code in {301, 302, 303, 307, 308}:
			return r.headers["Location"]
		else:
			raise NoRedirect("Unexpected status code: {}".format(r.status_code), response=r)

	except KeyError: # Location not provided. is this really raised?
		raise HTTPError("Location header not found", response=r)

class URLRequest(object):

	def __init__(self, url, headers=None, timeout=120, cookiejar=None, basicauth=None, openfunc=None):
		# type: (str, Optional[dict], Optional[float], Optional[CookieJar], Optional[Tuple[str, str, str]], Optional[Callable]) -> None

		self.url = url
		self.timeout = timeout
		headers = headers or {}

		if not openfunc:
			openfunc = URLRequestBuilder(cookiejar=None, basicauth=None).getfunc()

		req = request.Request(url, data=None, headers=headers)
		self.response = openfunc(req, timeout=timeout)
		self.headers = self.response.info() # type: HTTPMessage

	def _close(self): # unused
		self.response.close()

	def __enter__(self):
		# type: () -> URLRequest

		return self

	def __exit__(self, exc_type, exc_value, traceback):
		self.response.close()

	def _content_length(self):
		# type: () -> Optional[int]

		try:
			content_length = self.headers["Content-Length"]
			return int(content_length)
		except (KeyError, ValueError, TypeError):
			return None

	def _last_modified(self):
		# type: () -> Optional[float]

		try:
			last_modified = self.headers["Last-Modified"]
			return parsedate_to_timestamp(last_modified)
		except (KeyError, TypeError, ValueError):
			return None

	def get_redirect_url(self):
		self.response.close()
		return self.response.geturl()

	def _load(self):
		# type: () -> bytes

		with FileLike(self) as fp:
			try:
				return fp.read()
			except URLError:
				raise TimeOut("Timed out after {}s".format(self.timeout), response=self.response)

	def _json(self):
		# type: () -> dict

		with FileLike(self) as fp:
			try:
				return json.load(fp)
			except URLError:
				raise TimeOut("Timed out after {}s".format(self.timeout), response=self.response)

	def _download(self, basepath, filename=None, fn_prio=None, overwrite=False, suffix=".partial", report=None):
		# type: (str, Optional[str], Optional[Tuple[int, int, int, int]], bool, str, Optional[Callable[[int, int], None]]) -> Tuple[Optional[int], str]

		"""PermissionError: [WinError 32] Der Prozess kann nicht auf die Datei zugreifen,
			da sie von einem anderen Prozess verwendet wird:
			'G:\\PUBLIC\\Audio\\Podcasts\\Radio Nukular\\radio_nukular_057.m4a.partial'
			-> 'G:\\PUBLIC\\Audio\\Podcasts\\Radio Nukular\\radio_nukular_057.m4a'
		"""

		logger.debug("Downloading %s to %s", self.url, basepath)

		if fn_prio is None:
			fn_prio = (0, 1, 2, 3)

		filenames = {
			0: filename,
			1: get_filename(self.headers),
			2: get_filename_from_url(self.response.geturl()), # url after redirect
			3: get_filename_from_url(self.url),
		}

		filenames = [filenames[p] for p in fn_prio]
		logger.info("Filenames: {}".format(", ".join(map(str, filenames))))

		filename = first_not_none(filenames)

		if not filename:
			raise ValueError("Please provide a filename")

		filename = safe_filename(filename)
		fullpath = os.path.join(basepath, filename)
		tmppath = fullpath + suffix

		if not overwrite and os.path.exists(fullpath):
			raise FileExistsError(errno.EEXIST, "File already exists", fullpath)

		try:
			os.makedirs(basepath) # exist_ok=True is new in 3.2
		except OSError:
			pass

		content_length = self._content_length()

		try:
			# actually download the file
			with open(tmppath, "wb") as out:
				# https://www.ietf.org/mail-archive/web/httpbisa/current/msg27484.html
				transferred = copyfilelike(self.response, out, content_length, report=report)

		except (socket.timeout, URLError):
			logger.warning("Timeout after {}s at {}: {}".format(self.timeout, self.response.geturl(), self.headers))
			raise TimeOut("Timed out after {}s".format(self.timeout), response=self.response)

		except ConnectionResetError as e:
			logger.warning("Connection was reset during download: %s", str(e))
			raise DownloadInterrupted("Connection was reset during download")

		except Exception as e:
			logger.exception("Downloading failed mid file. Handle.")
			raise

		if content_length and content_length != transferred:
			logger.info("{} {}".format(content_length, transferred))
			raise ContentInvalidLength(tmppath, content_length, transferred)

		last_modified = self._last_modified()
		if last_modified:
			os.utime(tmppath, (-1, last_modified))

		os.rename(tmppath, fullpath)

		return (content_length, filename)

	def load(self):
		# type: () -> bytes

		try:
			return self._load()
		finally:
			self.response.close()

	def json(self):
		# type: () -> bytes

		try:
			return self._json()
		finally:
			self.response.close()

	def download(self, basepath, filename=None, fn_prio=None, overwrite=False, suffix=".partial", report=None):
		# type: () -> Tuple[Optional[int], str]

		try:
			return self._download(basepath, filename, fn_prio, overwrite, suffix, report)
		finally:
			self.response.close()

if __name__ == "__main__":
	from argparse import ArgumentParser

	parser = ArgumentParser(description="Downloads url to file")
	parser.add_argument("url", help="url to download")
	parser.add_argument("dir", help="output directory")
	parser.add_argument("-f", "--filename", help="filename")
	args = parser.parse_args()

	URLRequest(args.url).download(args.dir, args.filename)



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map, range
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Iterator, Tuple

def obj2tuple(obj):
	return (obj, )

def row_indices(n, square_length):
	""" Returns the one-dimensional indices of all the cells belonging to the
		same row as index `n`.
	"""

	j = n // square_length
	return range(j * square_length, (j + 1) * square_length)

def col_indices(n, square_length):
	""" Returns the one-dimensional indices of all the cells belonging to the
		same column as index `n`.
	"""

	j = n % square_length
	return range(j, square_length * (square_length - 1) + j + 1, square_length)

def subblock_indices(n, outer_square_length, inner_square_length):
	# type: (int, int, int) -> Iterator[Tuple[int, int]]

	""" Returns the one-dimensional indices of all the cells belonging to the
		same block as index `n`.
		`outer_square_length` is the number of cells per block.
		`inner_square_length` current has to be sqrt(block_size).
	"""

	i = outer_square_length
	j = inner_square_length

	if n >= i * i:
		raise ValueError("n not contained in matrix")

	if j * j != i:
		raise ValueError("Currently only squares are supported")

	x, y = (n % i // j, n // i // j)

	for iy in range(j):
		for ix in range(j):
			yield (y * i * j) + (i * iy) + (x * j) + (ix)

def triangular_indices(n):

	""" Returns all combinations of indices for a sequence of length `n` with duplicates and self-combinations.
		triangular_indices(4) -> (0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)
	"""

	for i in range(n-1):
		for j in range(i+1, n):
			yield i, j

def to_2d_index(n, width):
	return n // width, n % width

def indices_2d(n, width):
	# type: (int, int) -> Iterator[Tuple[int, int]]

	for i in range(n):
		yield i // width, i % width

def window_combinations_indices(size, window_size):
	# type: (int, int) -> Iterator[Tuple[int, int]]

	""" Yields all pairs of indices of a sequence of length `size` which are within a window
		of size `window_size`.

		Example:
			window_combinations_indices(3, 2) -> (0, 1), (1, 2)
			window_combinations_indices(4, 3) -> (0, 1), (0, 2), (1, 2), (1, 3), (2, 3)
	"""

	if window_size > size:
		raise ValueError("window size cannot exceed size")

	for a in range(0, size - 1):
		end = min(a + window_size, size)
		for b in range(a+1, end):
			yield a, b

def _combination_indices(start, sizes):

	""" earlier (left) dimensions can never be higher than later (right) dimensions. """

	if len(sizes) > 1:
		for i in range(start, sizes[0]):
			for j in _combination_indices(i, sizes[1:]):
				yield (i, ) + j
	else:
		for i in map(obj2tuple, range(start, sizes[0])):
			yield i

def combination_indices(*sizes):

	""" Given sequences of size a, b, c, ... yield the indices to index into each of them,
		so that all combinations of elements from the sequences are returned.
	"""

	return _combination_indices(0, sizes)



from __future__ import absolute_import, division, print_function, unicode_literals

import re

_markdown_newline_pat = re.compile(r"[^\n]\n[^\n]")
_markdown_newline_func = lambda x: x.group(0).replace("\n", "  \n")

def markdown_newline(s):
	# type: (str, ) -> str

	""" Convert newlines to markdown linebreaks. """

	return _markdown_newline_pat.sub(_markdown_newline_func, s)

_markdown_urls_pat = re.compile(r"(?<![\<])((?:http|https|ftp):\/\/[a-zA-Z0-9\-\.]+\.[a-zA-Z0-9]{2,}(?:\/[\:\/\?\#\[\]\@\!\$\&\'\(\)\*\+\;\=a-zA-Z0-9\-\._\~\%]*)?)(?![\>])")

def markdown_urls(s, ignore_trailing_dot=True):
	# type: (str, ) -> str

	""" Converts text with URLs to text with markdown formatted URLs.
		markdown_urls("Visit https://google.com") -> "Visit <https://google.com>"
		This function is idempotent, which means that it will not convert URLs which are already in the correct format.
		markdown_urls("Visit <https://google.com>") -> "Visit <https://google.com>"
	"""

	def markdown_dotfix(m):
		url, = m.groups()
		if ignore_trailing_dot and url.endswith("."):
			return "<{}>.".format(url[:-1])
		else:
			return "<{}>".format(url)

	# return _markdown_urls_pat.sub(r"<\1>", s)
	return _markdown_urls_pat.sub(markdown_dotfix, s)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewvalues
import sys
from time import sleep
from collections.abc import Sequence
from typing import TYPE_CHECKING

from .dict import update
from .exceptions import InconsistentState, DownloadFailed, WouldBlockForever, ExternalProcedureUnavailable, assert_type

from aria2p import Client
from aria2p.client import DEFAULT_HOST, DEFAULT_PORT, ClientException
from requests.exceptions import ConnectionError

if TYPE_CHECKING:
	from typing import Any, Optional, Tuple, TextIO, List

def aria_bool(value):
	# type: (Optional[bool], ) -> Optional[str]

	""" aria2 rpc requires strings for some boolean arguments like 'continue'
	"""

	if value is None:
		return None
	elif value is True:
		return "true"
	elif value is False:
		return "false"
	else:
		raise ValueError(str(value))

class AriaDownloader(object):

	default_global_options = {
		"max-concurrent-downloads": 5,
		"remote-time": True,
	}

	default_options = {
		"max-connection-per-server": 1,
		"split": 5,
		"always-resume": True,
	}

	max_num_results = 100

	""" Download manager which uses aria2 instance to actually download the files.
		Tries to respect other users of the same instance and doesn't interfere with them.
	"""

	def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT, secret="", poll=1.0, global_options=None):  # nosec
		# type: (str, int, str, float, Optional[dict]) -> None

		""" Initialize the aria2 client with `host`, `port` and `secret`.
			Poll aria2 every `poll` seconds.
		"""

		self.aria2 = Client(host, port, secret)
		self.poll = poll

		if global_options:
			global_options = self.default_global_options.copy()
			global_options.update(global_options)
			self.aria2.change_global_option(global_options)
		else:
			self.aria2.change_global_option(self.default_global_options)

		self.gids = set()

	def query(self, method, *args, **kwargs):

		try:
			return getattr(self.aria2, method)(*args, **kwargs)
		except ClientException as e:
			if e.code == 1:
				# either our code is bad, or some external actor removed our gid from aria
				raise InconsistentState(e.message)
			else:
				raise
		except ConnectionError as e:
			raise ExternalProcedureUnavailable(e)

	def pause_all(self):
		# type: () -> None

		for gid in self.gids:
			self.query("pause", gid)

	def resume_all(self):
		# type: () -> None

		for gid in self.gids:
			self.query("unpause", gid)

	def _entries(self, entries):
		return {entry["gid"]: entry for entry in entries if entry["gid"] in self.gids}

	def managed_downloads(self):
		# type: () -> int

		return len(self.gids)

	def block_one(self, progress_file=sys.stdout):
		# type: (Optional[TextIO], ) -> Tuple[str, str]

		""" Blocks until one download is done. """

		while True:
			""" fixme: This loop has a (not very serious) race condition.

				For example a download might change its status from waiting to active
				during the two queries. Then it would not be found.
				This cannot be fixed by changing the query order as the status change graph
				(waiting<->active->stopped) cannot be sorted topologically.
				And even if you could, returning stopped downloads has priority over
				waiting/active ones, as otherwise we would block for too long.

				Does using a multi/batch-call fix this?
			"""

			entries = self._entries(self.query("tell_stopped", 0, self.max_num_results))  # complete or error
			if entries:
				gid, entry = entries.popitem()

				try:
					if entry["status"] == "complete":
						assert len(entry["files"]) == 1
						return gid, entry["files"][0]["path"]
					elif entry["status"] == "error":
						raise DownloadFailed(gid, entry["errorCode"], entry["errorMessage"])
					else:
						raise RuntimeError("Unexpected status: {}".format(entry["status"]))

				finally:
					self.remove_stopped(gid)

			entries = self._entries(self.query("tell_active"))  # active
			if entries:
				if progress_file:
					completed = sum(int(entry["completedLength"]) for entry in viewvalues(entries))
					total = sum(int(entry["totalLength"]) for entry in viewvalues(entries))
					speed = sum(int(entry["downloadSpeed"]) for entry in viewvalues(entries))
					print("{} downloads: {}/{} bytes {} bytes/sec".format(len(entries), completed, total, speed), file=progress_file, end="\r")

				sleep(self.poll)
				continue

			entries = self._entries(self.query("tell_waiting", 0, self.max_num_results)) # waiting or paused
			if entries:
				print("{} downloads waiting or paused".format(len(entries)), end="\r")
				sleep(self.poll)
				continue

			if self.gids:
				""" Actually only this check is sensitive to race condition, as the looping logic
					would care of retrying otherwise.
					However this is the only way to check for external modifications.
				"""
				raise InconsistentState("Some downloads got lost. We either encoutered a race condition \
					or some external actor removed the download")

			raise WouldBlockForever("No downloads active or waiting")

	def block_all(self):
		# type: () -> List[Tuple[Any, Any]]

		ret = []  # type: Tuple[Any, Any]

		while True:
			try:
				gid, path = self.block_one()
				ret.append((None, path))
			except WouldBlockForever:
				break
			except DownloadFailed as e:
				ret.append((e.args, None))

		return ret

	def remove_stopped(self, gid):
		# type: (str, ) -> None

		self.gids.remove(gid)

		# removes a complete/error/removed download
		# fails on active/waiting/paused (on CTRL-C for example)
		self.query("remove_download_result", gid)

	def block_gid(self, gid, progress_file=sys.stdout):
		# type: (str, Optional[TextIO]) -> str

		""" Blocks until download is done.
			If progress printing is not needed, `progress_file` should be set to `None`.
			Returns the path of the downloaded file on disk.
		"""

		if gid not in self.gids:
			raise KeyError("Invalid GID")

		try:
			while True:
				s = self.query("tell_status", gid)

				status = s["status"]

				if status == "active":
					if progress_file:
						print(s["completedLength"], "/", s["totalLength"], "bytes", s["downloadSpeed"], "bytes/sec", file=progress_file, end="\r")

				elif status == "waiting":
					if progress_file:
						print("waiting", file=progress_file, end="\r")

				elif status == "paused":
					if progress_file:
						print("paused", file=progress_file, end="\r")

				elif status == "error":
					raise DownloadFailed(gid, s["errorCode"], s["errorMessage"]) # RuntimeError: No URI available. errorCode=8 handle

				elif status == "complete":
					assert len(s["files"]) == 1
					return s["files"][0]["path"]

				elif status == "removed":
					raise RuntimeError("Someone removed our download...")

				else:
					raise RuntimeError("Unexpected status: {}".format(status))

				sleep(self.poll)
		finally:
			self.remove_stopped(gid)

	def download(self, uri, path=None, filename=None, headers=None, max_connections=None, split=None, continue_=None, retry_wait=None):
		# type: (str, Optional[str], Optional[str], Optional[Sequence[str]], Optional[int], Optional[int], Optional[bool], Optional[int]) -> str

		""" Downloads `uri` to directory `path`.
			Does not block. Returns a download identifier.
		"""

		if headers is not None:
			assert_type("headers", headers, Sequence)

		options = self.default_options.copy()
		update(options, {
			"dir": path,
			"out": filename,
			"max-connection-per-server": max_connections,
			"split": split,
			"continue": aria_bool(continue_),
			"retry-wait": retry_wait,
			"header": headers,
		})

		gid = self.query("add_uri", [uri], options)
		self.gids.add(gid)
		return gid

	def download_x(self, num, uri, path=None, filename=None, headers=None, max_connections=None, split=None, continue_=None, retry_wait=None):
		# type: (int, str, Optional[str], Optional[str], Optional[Sequence[str]], Optional[int], Optional[int], Optional[bool], Optional[int]) -> Optional[Tuple[str, str, str]]

		queued_gid = self.download(uri, path, filename, headers, max_connections, split, continue_, retry_wait)
		if self.managed_downloads() >= num:
			finished_gid, path = self.block_one()
			return queued_gid, finished_gid, path

		return None

class DownloadManager(object):

	""" To use from a single thread.

		State transitions:
		waiting -> running (automatically)
		running -> paused (manually)
		paused -> running (manually)
		running -> complete (automatically)
		running -> error (automatically)
	"""

	def __init__(self):
		pass

	def download_with_callback(self, url, path=None, filename=None, headers=None, force=False, func=None):
		""" Starts or enqueues download and a unique id for that download job.
			if `force` is True, the download is started without regard for the current number
			of downloads and queue position.
		"""

	def block_uid(self, uid):
		""" Blocks until `uid` completes or errors
			and returns information about that download.
		"""

	def block_one(self):
		""" Block until any download completes or errors
			and returns information about that download.
		"""

	def block_active(self, x):
		""" Block if more than `x` downloads are currently active
			until x or less are active. Yields information about completed/errored downloads
			as they become available.
		"""

	def cancel(self, uid, states=None):
		""" Cancels a download identified by `uid`. if `states` is None, it will be
			- removed from the queue if waiting
			- canceled if downloading
			- removed if completed
			- removed if errored.
			The previous state will be returned,
		"""

		# sets the status if `uid` to canceled.

	def clean(self, uid, states=None):
		""" Removed all files associated to `uid` from disk. `states` is a list of acceptable states.
			For example ("stopped", "error") will only remove files if the download
			was stopped mid-progress or because of an error.
			Take notice that ("complete") will remove files of completed downloads from disk.
		"""

		# doesn't change status

	def forget(self, uid):
		""" Remove `uid` from the manager completely. It can only do so for downloads
			with the state "canceled".
		"""

	# for convenience only

	def block_all(self, uid):
		return self.block_active(0)

	def download_and_block(self, url):
		uid = self.download_with_callback(url)
		return self.block_uid(uid)

if __name__ == "__main__":

	from argparse import ArgumentParser
	parser = ArgumentParser()
	parser.add_argument("uris", metavar="URI", nargs="+", help="URLs to download")
	parser.add_argument("--outpath", default=".", help="Output directory")
	parser.add_argument("--max", default=2, type=int, help="Maximum concurrent downloads")
	args = parser.parse_args()

	d = AriaDownloader()
	for uri in args.uris:
		path = d.download_x(args.max, uri, args.path)
		print("Downloaded {} to {}".format(uri, path))

	for a in d.block_all():
		print(a)



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range
from operator import itemgetter
from typing import TYPE_CHECKING

from .func import identity
from .indexing import triangular_indices

if TYPE_CHECKING:
	from typing import Callable, Iterable, Iterator, MutableSequence, Tuple, Sequence, TypeVar
	from .typing import Comparable
	T = TypeVar("T")

class LazyStringList(object):

	def __init__(self, length):
		# type: (int, ) -> None

		self.length = length

	def __getitem__(self, idx):
		# type: (int, ) -> str

		return str(idx)

	def __len__(self):
		# type: () -> int

		return self.length

	def __iter__(self):
		# type: () -> Iterator[str]

		return map(str, range(self.length))

	def __contains__(self, idx):
		# type: (int, ) -> bool

		return 0 <= idx < self.length

def pop_many(seq, func):
	# type: (MutableSequence[T], Callable) -> Iterator[T]

	""" `pop()`s values from `seq` where func(value) is true.
		For performance reasons the elements are popped
		and yielded in reverse order.
	"""

	idx = [i for i, elm in enumerate(seq) if func(elm)]

	for i in reversed(idx):
		yield seq.pop(i)

def triangular(seq):
	# type: (Sequence[T], ) -> Iterator[Tuple[T, T]]

	""" Returns all combinations of items of `seq` with duplicates and self-combinations.
		triangular([1, 2, 3, 4]) -> (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)
	"""

	for i, j in triangular_indices(len(seq)):
		yield seq[i], seq[j]

def sliding_window(seq, size, step=1):
	# type: (Sequence[T], int, int) -> Iterator[Sequence[T]]

	""" Similar to `iter.x_wise()` except that it returns slices of a sequence.
		sliding_window([1, 2, 3], 2) -> [1, 2], [2, 3]
		sliding_window([1, 2, 3, 4, 5], 2, 3) -> [1, 2], [4, 5]
	"""

	seqlen = len(seq)
	if size > seqlen:
		raise ValueError("size must be less or equal the length of the sequence")

	for i in range(0, seqlen-size+1, step):
		yield seq[i:i+size]

def batch(seq, size):
	# type: (Sequence[T], int) -> Iterator[Sequence[T]]

	""" Similar to `iter.batch()` except that it returns slices of a sequence.
		sliding_window([1, 2, 3], 2) -> [1, 2], [2, 3]
		sliding_window([1, 2, 3, 4, 5], 2, 3) -> [1, 2], [4, 5]
	"""

	if size < 1:
		raise ValueError("size must be larger than 0")

	seqlen = len(seq)

	for i in range(0, (seqlen + size - 1) // size):
		yield seq[i*size:(i+1)*size]

def delete_duplicates_from_sorted_sequence(seq, key=None):
	# type: (MutableSequence[T], Optional[Callable[[T], Comparable]) -> None

	""" Deletes duplicates from a sorted sequence `seq` based on `key`.
		Works in-place.
	"""

	key = key or identity
	i = 0
	while i < len(seq) - 1:
		if key(seq[i]) == key(seq[i+1]):
			del seq[i+1]
		else:
			i += 1

remove_duplicate_rows_from_sorted_table = lambda seq, key: delete_duplicates_from_sorted_sequence(seq, itemgetter(key))

def cycle_sort(seq):
	# type: (MutableSequence[T], ) -> Iterator[Tuple[int, int]]

	""" Sort a sequence in place and yield swaps.
		based on: https://rosettacode.org/wiki/Sorting_algorithms/Cycle_sort#Python
	"""

	# Loop through the sequence to find cycles to rotate.
	for cycle_start, item in enumerate(seq):

		# Find where to put the item.
		pos = cycle_start
		for item2 in seq[cycle_start + 1:]:
			if item2 < item:
				pos += 1

		# If the item is already there, this is not a cycle.
		if pos == cycle_start:
			continue

		# Otherwise, put the item there or right after any duplicates.
		while item == seq[pos]:
			pos += 1
		seq[pos], item = item, seq[pos]
		yield cycle_start, pos

		# Rotate the rest of the cycle.
		while pos != cycle_start:

			# Find where to put the item.
			pos = cycle_start
			for item2 in seq[cycle_start + 1:]:
				if item2 < item:
					pos += 1

			# Put the item there or right after any duplicates.
			while item == seq[pos]:
				pos += 1
			seq[pos], item = item, seq[pos]
			if pos != cycle_start:
				yield cycle_start, pos



from typing import TYPE_CHECKING

import torch
from torch.distributions.categorical import Categorical

if TYPE_CHECKING:
	from torch import tensor

def categorical(probs, size=torch.Size()):
	# type: (tensor, tensor) -> tensor

	return Categorical(probs).sample(size)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import PY2

import csv
from itertools import islice
from functools import partial
from io import open
from typing import TYPE_CHECKING

from .dict import itemgetter
from .func import identity, zipmap, compose

if TYPE_CHECKING:
	from typing import Iterator, List, Optional, Sequence, Callable

def iter_csv(path, delimiter=",", encoding="utf-8"):
	# type: (str, str, str, str, str) -> Iterator[List[str]]

	if PY2:
		with open(path, "rb") as fr:
			for data in csv.reader(fr, delimiter=delimiter.encode(encoding)):
				yield list(field.decode(encoding) for field in data)
	else:
		with open(path, "rt", encoding="utf-8", newline="") as fr:
			for data in csv.reader(fr, delimiter=delimiter):
				yield data

def read_csv(path, delimiter=",", skip=0, usecols=None, dtype=None, encoding="utf-8"):
	# type: (str, str, int, Optional[Sequence[int]], Optional[Sequence[Callable]], str) -> Iterator[tuple]

	with open(path, "rt", encoding=encoding, newline="") as csvfile:
		if usecols:
			if dtype:
				getcols = compose(partial(zipmap, dtype), itemgetter(usecols))
			else:
				getcols = itemgetter(usecols)
		else:
			if dtype:
				getcols = partial(zipmap, dtype)
			else:
				getcols = identity

		for row in islice(csv.reader(csvfile, delimiter=delimiter), skip, None):
			print(row)
			yield tuple(getcols(row))



from __future__ import absolute_import, division, print_function, unicode_literals

from io import open

from lxml import etree  # nosec

def xml_xslt_to_xhtml(path_xml, path_xslt, path_xhtml):

	""" Only use with trusted xml data """

	xml = etree.parse(path_xml)  # nosec
	xslt = etree.parse(path_xslt)  # nosec
	transform = etree.XSLT(xslt)
	newdom = transform(xml)
	with open(path_xhtml, "wb") as xhtml:
		newdom.write(xhtml, pretty_print=True, xml_declaration=True, encoding="utf-8")



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from PIL import Image, ImageDraw, ImageFont
from PIL.ExifTags import TAGS, GPSTAGS
import piexif

from .exceptions import NoActionNeeded

if TYPE_CHECKING:
	from typing import Any, Tuple, Union, Dict
	Color = Union[str, Tuple[int, int, int]]

def multiline_textsize(text, ttf, spacing=4):
	# type: (str, ImageFont, int) -> Tuple[int, int]

	lines = text.splitlines()

	width = 0
	height = (len(lines) - 1) * spacing

	for line in lines:
		w, h = ttf.getsize(line)
		width = max(width, w)
		height = height + h

	return width, height

def exifinfo(image):
	# type: (Image, ) -> Dict[str, Any]

	ret = dict()

	exifd = image._getexif()

	if exifd is None:
		return dict()

	for k, v in exifd.items():
		try:
			tag = TAGS[k]
			if tag == "GPSInfo":
				ret[tag] = dict()
				for k, v in v.items():
					try:
						gpstag = GPSTAGS[k]
						ret[tag][gpstag] = v
					except KeyError:
						pass
			else:
				ret[tag] = v
		except KeyError:
			pass

	return ret

def text_with_outline(draw, pos, text, font, fillcolor, outlinecolor, outlinesize=1):
	# type: (ImageDraw, Tuple[int, int], str, ImageFont, Color, Color) -> None

	x, y = pos
	delta = outlinesize

	# thin border
	draw.text((x-delta, y), text, font=font, fill=outlinecolor)
	draw.text((x+delta, y), text, font=font, fill=outlinecolor)
	draw.text((x, y-delta), text, font=font, fill=outlinecolor)
	draw.text((x, y+delta), text, font=font, fill=outlinecolor)

	# thicker border
	draw.text((x-delta, y-delta), text, font=font, fill=outlinecolor)
	draw.text((x+delta, y-delta), text, font=font, fill=outlinecolor)
	draw.text((x-delta, y+delta), text, font=font, fill=outlinecolor)
	draw.text((x+delta, y+delta), text, font=font, fill=outlinecolor)

	# now draw the text over it
	draw.text((x, y), text, font=font, fill=fillcolor)

def write_text(img, text, alignment="TL", fillcolor=(255, 255, 255), outlinecolor=(0, 0, 0), fontratio=0.03, padding=(5, 5)):
	# (Image, str, str, Color, Color, float, Tuple[int, int]) -> None

	if alignment not in {"TL", "TC", "TR", "BL", "BC", "BR"}:
		raise ValueError("Invalid alignment: {}".format(alignment))

	fontsize = int(img.height * fontratio)
	font = ImageFont.truetype("arial.ttf", fontsize)

	d = ImageDraw.Draw(img)
	size_text = d.textsize(text, font=font)

	if alignment == "TL":
		pos = padding
	elif alignment == "TC":
		pos = (img.width // 2 - size_text[0] // 2, padding[1])
	elif alignment == "TR":
		pos = (img.width - padding[0] - size_text[0], padding[1])
	elif alignment == "BL":
		pos = (padding[0], img.height - padding[1] - size_text[1])
	elif alignment == "BC":
		pos = (img.width // 2 - size_text[0] // 2, img.height - padding[1] - size_text[1])
	elif alignment == "BR":
		pos = (img.width - padding[0] - size_text[0], img.height - padding[1] - size_text[1])

	text_with_outline(d, pos, text, font, fillcolor, outlinecolor, 2)

def fix_orientation(img, exif):
	# type (Image, dict) -> dict

	orientation = exif["0th"][piexif.ImageIFD.Orientation]

	if orientation == 1:
		raise NoActionNeeded("File already properly rotated")
	elif orientation == 2:
		img = img.transpose(Image.FLIP_LEFT_RIGHT)
	elif orientation == 3:
		img = img.transpose(Image.ROTATE_180)
	elif orientation == 4:
		img = img.transpose(Image.FLIP_TOP_BOTTOM)
	elif orientation == 5:
		img = img.transpose(Image.TRANSPOSE)
	elif orientation == 6:
		img = img.transpose(Image.ROTATE_270)
	elif orientation == 7:
		img = img.transpose(Image.TRANSVERSE)
	elif orientation == 8:
		img = img.transpose(Image.ROTATE_90)
	else:
		raise ValueError("Unsupported orientation")

	exif["0th"][piexif.ImageIFD.Orientation] = 1
	return img, exif




from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

import numpy as np

from .numpy import unblock, bincount_batch

if TYPE_CHECKING:
	from typing import Union, Tuple
	from fractions import Fraction
	Rational = Union[float, Fraction]

def grayscale(arr):
	return np.sum(arr, axis=-1) // arr.shape[-1]

def histogram_1d(arr, levels):
	# type: (np.ndarray, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., x]. Histogrammed over x and batched over the remaining dimensions.
		Output shape: [batch..., levels]
	"""

	return bincount_batch(arr, -1, levels)

def resize_oar(max_width, max_height, dar):
	# type: (int, int, Rational) -> Tuple[int, int]

	maxdar = max_width / max_height

	if dar >= maxdar:  # wider than it should be
		width = max_width
		height = int(width / dar)
	else:  # thinner than it should be
		height = max_height
		width = int(height * dar)

	return width, height

def resize_maxsize(max_width, max_height, width, height):
	# type: (int, int, int, int) -> Tuple[int, int]

	return resize_oar(max_width, max_height, width / height)

def histogram_2d(arr, levels):
	# type: (np.ndarray, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., y, x]. Histogrammed over x and y and batched over
		the remaining dimensions.
		Output shape: [batch..., levels]
	"""

	if len(arr.shape) < 2:
		raise ValueError("arr must be at least 2-dimensional")

	newshape = arr.shape[:-2] + (arr.shape[-2] * arr.shape[-1], )
	flattened = np.reshape(arr, newshape)

	return bincount_batch(flattened, -1, levels)

def block_histogram_2d(arr, by, bx, levels):
	# type: (np.ndarray, int, int, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., y, x]. Histogrammed over blocks of size bx and by
		and batched over the remaining dimensions.
		Output shape: [batch..., y/by, x/bx, levels]
	"""

	invx = arr.shape[-1] // bx  # dimensions in unblock go from innerst to outerst
	invy = arr.shape[-2] // by
	blocks = unblock(arr, invx, invy)
	block_hists = histogram_1d(blocks, levels)
	return block_hists.reshape(arr.shape[:-2] + (invy, invx, -1))

def image_histogram(arr, levels=256):
	# type: (np.ndarray, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., x, y, channel]. It is summed over channels to create a grayscale
		image, then histogrammed over x and y and batched over the remaining dimensions.
		Output shape: [batch..., levels]
	"""

	if len(arr.shape) < 3:
		raise ValueError("arr must be at least 3-dimensional")

	gray = grayscale(arr)
	return histogram_2d(gray, levels)

def image_block_histogram(arr, bx, by, levels=256):
	# type: (np.ndarray, int, int, int) -> np.ndarray

	""" Input shape of `arr`: [batch..., x, y, channels]. It is summed over channels to create a grayscale
		image, then histogrammed over x and y and batched over the remaining dimensions.
		Output shape: [batch..., bx, by, levels]
	"""

	if len(arr.shape) < 3:
		raise ValueError("arr must be at least 3-dimensional")

	gray = grayscale(arr)
	return block_histogram_2d(gray, bx, by, levels)



from __future__ import absolute_import, division, print_function, unicode_literals

""" Scripts using line_profiler.profile can import this to be able to run unmodified
	without the profiler present.
"""
import logging

try:
	profile = profile # bind so it can be imported
	logging.info("Running profiler")

except NameError:
	def profile(func):
		return func



from __future__ import absolute_import, division, print_function, unicode_literals

import logging, codecs, sqlite3
from typing import TYPE_CHECKING

from .iter import batch, progress
from .signal import safe_for_loop

if TYPE_CHECKING:
	from sqlite3 import Cursor, Iterable

def quote_identifier(s, errors="strict"):
	# type: (str, str) -> str

	""" Escapes sqlite identifiers (like table or column names).
		Copied from: https://stackoverflow.com/a/6701665
	"""

	encodable = s.encode("utf-8", errors).decode("utf-8")
	nul_index = encodable.find("\x00")

	if nul_index >= 0:
		error = UnicodeEncodeError("NUL-terminated utf-8", encodable, nul_index, nul_index + 1, "NUL not allowed")
		error_handler = codecs.lookup_error(errors)
		replacement, _ = error_handler(error)
		encodable = encodable.replace("\x00", replacement)

	return "\"" + encodable.replace("\"", "\"\"") + "\""

def vacuum(db_path):
	# type: (str, ) -> None

	""" Vacuums a sqlite database. """

	with sqlite3.connect(db_path) as conn:
		conn.execute("VACUUM")

def batch_executer(cursor, query_str, it, batch_size=10000, exclusive=True):
	# type: (Cursor, str, Iterable[tuple], int, bool) -> int

	""" Execute `query_str` with parameters from `it` batch-wise with batches of size `batch_size`.
		If `exclusive` is True the database will be locked in exclusive mode.
	"""

	if exclusive:
		cursor.execute("PRAGMA locking_mode=EXCLUSIVE") # might speed things up

	entries = 0

	for queries_batch in batch(progress(it), batch_size):

		# need to cache batch data, because if the iterable is exhausted,
		# executemany raises `sqlite3.ProgrammingError`
		data = list(queries_batch) 
		if data:
			cursor.execute("BEGIN TRANSACTION")
			cursor.executemany(query_str, data)
			cursor.execute("COMMIT TRANSACTION")
			entries += len(data)

	return entries

def safe_batch_executer(cursor, query_str, it, batch_size=10000, exclusive=True):
	# type: (Cursor, str, Iterable[tuple], int, bool) -> None

	""" Execute `query_str` with parameters from `it` batch-wise with batches of size `batch_size`.
		If `exclusive` is True the database will be locked in exclusive mode.
		If an db integrity error occurs, the batch will be skipped and the transaction for
		this batch rolled back.
		If the iterable is advanced, the user can be sure its elements will be inserted into the
		database, even if a KeyboardInterrupt is received in-between.
	"""

	if exclusive:
		cursor.execute("PRAGMA locking_mode=EXCLUSIVE")  # might speed things up

	source = batch(progress(it), batch_size)

	def sqlexec(queries_batch):
		try:
			cursor.execute("BEGIN TRANSACTION")
			cursor.executemany(query_str, queries_batch)
			cursor.execute("COMMIT TRANSACTION")
		except sqlite3.IntegrityError:
			logging.info("Skipping batch")
			cursor.execute("ROLLBACK TRANSACTION")
		except sqlite3.OperationalError:
			raise

	try:
		safe_for_loop(source, sqlexec)
	except KeyboardInterrupt:
		logging.info("Batch execution safely interrupted")



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

import av

if TYPE_CHECKING:
	from typing import Iterator
	import numpy as np

class CorruptFile(Exception):
	pass

def iter_video(path):
	# type: (str, ) -> Iterator[np.ndarray]

	""" Yield frames from a video file. For video device support see `genutility.cv.iter_video`.
	"""

	with av.open(path, "r") as container:

		vs = container.streams.video[0]
		vs.thread_type = "AUTO"

		for frame in container.decode(vs):
			if frame.is_corrupt:
				raise CorruptFile("Frame %s at %s of %s is corrupt", frame.index, frame.time, path)
			yield frame.to_ndarray() # or frame.to_rgb()



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from collections.abc import Iterable, Sized

class SizedIterable(Iterable, Sized):
	pass

if TYPE_CHECKING:

	from typing import Any, Optional, Sequence, Union
	from typing_extensions import Protocol

	Number = Union[int, float]

	class Comparable(Protocol):

		def __eq__(self, other):
			# type: (Any, ) -> bool
			...

		def __ne__(self, other):
			# type: (Any, ) -> bool
			...

	class Orderable(Protocol):

		def __lt__(self, other):
			# type: (Any, ) -> bool
			...

		def __gt__(self, other):
			# type: (Any, ) -> bool
			...

		def __le__(self, other):
			# type: (Any, ) -> bool
			...

		def __ge__(self, other):
			# type: (Any, ) -> bool
			...

	class Computable(Protocol):

		def __neg__(self, other):
			# type: (Any, ) -> Any
			...

		def __add__(self, other):
			# type: (Any, ) -> Any
			...

		def __sub__(self, other):
			# type: (Any, ) -> Any
			...

		def __mul__(self, other):
			# type: (Any, ) -> Any
			...

		def __truediv__(self, other):
			# type: (Any, ) -> Any
			...

		def __floordiv__(self, other):
			# type: (Any, ) -> Any
			...

		def __abs__(self, other):
			# type: (Any, ) -> Any
			...

	class MutableComputable(Protocol):

		def __neg__(self, other):
			# type: (Any, ) -> Any
			...

		def __add__(self, other):
			# type: (Any, ) -> Any
			...

		def __iadd__(self, other):
			# type: (Any, ) -> Any
			...

		def __sub__(self, other):
			# type: (Any, ) -> Any
			...

		def __isub__(self, other):
			# type: (Any, ) -> Any
			...

		def __mul__(self, other):
			# type: (Any, ) -> Any
			...

		def __imul__(self, other):
			# type: (Any, ) -> Any
			...

		def __truediv__(self, other):
			# type: (Any, ) -> Any
			...

		def __itruediv__(self, other):
			# type: (Any, ) -> Any
			...

		def __floordiv__(self, other):
			# type: (Any, ) -> Any
			...

		def __ifloordiv__(self, other):
			# type: (Any, ) -> Any
			...

	class Cursor(Protocol):

		def callproc(self, procname, parameters):
			# type: (str, Optional[Sequence[Any]]) -> None
			...

		def close(self):
			# type: () -> None
			...

		def execute(self, operation, parameters):
			# types: (str, Optional[Sequence[Any]]) -> None
			...

		def executemany(self, operation, seq_of_parameters):
			# types: (str, Sequence[Sequence[Any]]) -> None
			...

		def fetchone(self):
			# types: () -> Optional[Sequence[Any]]
			...

		def fetchmany(self, size):
			# types: (Optional[int], ) -> Sequence[Sequence[Any]]
			...

		def fetchall(self):
			# types: () -> Sequence[Sequence[Any]]
			...

		def nextset(self):
			# type: () -> Optional[bool]
			...

		def setinputsizes(self, sizes):
			# types: (Sequence[int], ) -> None
			...

		def setoutputsize(self, size, column):
			# type: (int, Optional[int]) -> None
			...

	class Connection(Protocol):

		def close(self):
			# type: () -> None
			...

		def commit(self):
			# type: () -> None
			...

		def rollback(self):
			# type: () -> None
			...

		def cursor():
			# type: () -> Cursor
			...



from __future__ import absolute_import, division, print_function, unicode_literals

import re
from string import whitespace
from typing import TYPE_CHECKING

from .url import valid_uri_characters, get_url_pattern
from .string import replace_multiple
from .iter import collapse_any, collapse_all
from .unicode import unicode_categories

if TYPE_CHECKING:
	from typing import Iterable, Iterator, Union, Callable, Optional, Sequence

uni_cats = unicode_categories()

def extract_urls(text):
	# type: (str, ) -> Iterator[str]

	""" Yields URLs from `text`. """

	pattern = r"https?\:\/\/[" + re.escape(valid_uri_characters) + r"]+"
	for match in re.finditer(pattern, text):
		url = match.group(0)
		yield url

def replace_typographical_punctuation(s):
	# type: (str, ) -> str

	""" Replaces typographical punctuation with ASCII punctuation. """

	d = {
		"\N{Single Low-9 Quotation Mark}": "'",
		"\N{Double Low-9 Quotation Mark}": '"',
		"\N{Left Single Quotation Mark}": "'",
		"\N{Right Single Quotation Mark}": "'",
		"\N{Left Double Quotation Mark}": '"',
		"\N{Right Double Quotation Mark}": '"',
		"\N{Figure Dash}": "-",
		"\N{En Dash}": "-",
		"\N{Em Dash}": "--",
		"\N{Horizontal Bar}": "--",
		"\N{Double Low Line}": "_",
		"\N{Horizontal Ellipsis}": "...",
	}
	return replace_multiple(s, d)

def newlines_to_spaces(s):
	# type: (str, ) -> str

	""" Replaces newline characters with spaces. """

	d = {
		"\r": "",
		"\n": " ",
		"\N{NEXT LINE}": " ",
	}
	return replace_multiple(s, d)

def collapse_whitespace(s):
	# type: (str, ) -> str

	""" Collapses repeated whitespace into a single space character. """

	# separator_cats = ("Zl", "Zp", "Zs")
	# cats = tuple(uni_cats[i] for i in separator_cats)
	# separators = set.union(*cats)
	return "".join(collapse_all(s, set(whitespace), " "))

def collapse_space(s, space=" "):
	# type: (str, str) -> str

	""" Collapses repeated spaces into a single space character. """

	# slower: variant of this with join
	# slower: re.compile(r" {2,}").sub(" ", s)
	# slower: "".join(collapse_all(s, {" "}, " "))

	if s == space:
		return s

	ss = s.split(space)

	beg = space if not ss[0]  else ""
	end = space if not ss[-1] else ""

	return beg + space.join(i for i in ss if i) + end

def collapse_punctuation_symbols(s):
	# type: (str, ) -> str

	""" Collapses any successive Unicode punctuation symbol in `s`.
	"""

	punctuation_cats = ("Pc", "Pd", "Pe", "Pf", "Pi", "Po", "Ps")
	symbol_cats = ("Sc", "Sk", "Sm", "So")

	cats = tuple(uni_cats[i] for i in punctuation_cats+symbol_cats)
	unwantedset = set.union(*cats)
	return "".join(collapse_any(s, unwantedset))

class ReplaceURLs(object):

	def __init__(self, replacement, schemes=None):
		# type: (Union[str, Iterable], Optional[Sequence[str]]) -> None

		if isinstance(replacement, str):
			self.repl = replacement # type: Union[str, Callable[[str], str]]
		else:
			self.repl = lambda match: next(replacement) # type: ignore

		self.pattern = get_url_pattern(schemes)

	def __call__(self, s, count=0):
		# type: (str, int) -> str

		return self.pattern.sub(self.repl, s, count) # type: ignore # mypy error, `sub()` supports functions as `repl`



from __future__ import absolute_import, division, print_function, unicode_literals

# don't do `from builtins import str`
from setuptools import setup
from io import open
from itertools import chain

# nltk.download("punkt")

with open("README.md", "r", encoding="utf-8") as fr:
	long_description = fr.read()

extras_require = {
	"args": [],
	"algorithms": [],
	"atomic": [],
	"binary": [],
	"concurrency": [],
	"constants_physics": [],
	"constants_video": [],
	"csv": [],
	"datetime": [],
	"debug": [],
	"dict": [],
	"datastructures": [],
	"error": [],
	"exceptions": [],
	"factorial": [],
	"func": [],
	"file": [],
	"filesystem": [],
	"geometry": [],
	"hash": [],
	"iter": [],
	"imdb": [],
	"indexing": [],
	"latex": [],
	"math": [],
	"mongo": [],
	"object": [],
	"ops": [],
	"os": ["ctypes-windows-sdk>=0.0.7; sys_platform == 'win32'"],
	"pickle": [],
	"profile": [],
	"rand": [],
	"ringlist": [],
	"search": [],
	"sequence": [],
	"set": [],
	"signal": [],
	"sort": [],
	"sql": [],
	"statistics": [],
	"stdio": [],
	"string": [],
	"sqlite": [],
	"test": [],
	"time": [],
	"typing": [],
	"text": [],
	"twitch": [],
	"url": [],

	# "browscap": ["re2"], # unfinished
	"cholesky": ["numpy", "numba"],
	"config": ["jsonschema", "sortedcontainers"],
	"downloadmanager": ["aiohttp>=0.16.3", "orderedset"],
	"encoder": ["numpy", "nltk; python_version>='3'", "nltk<3.5; python_version<'3'"],
	"fingerprinting": ["numpy", "numba", "pillow"],
	"html": ["beautifulsoup4"],
	"http": ["requests"],
	"http2": ["hyper"],
	"image": ["numpy"],
	"json": ["jsonschema"],
	"lda": ["numpy", "nltk; python_version>='3'", "nltk<3.5; python_version<'3'"],
	"metrics": ["numpy"],
	"mp4": ["setuptools"],
	"multihash": ["rhash"],
	"net": ["netifaces"],
	"nlp": ["nltk; python_version>='3'", "nltk<3.5; python_version<'3'"],
	"regression": ["numpy"],
	"resizing": ["numpy", "opencv-python", "scikit-learn"],
	"scene_change_detection": ["numpy"],
	"sudoku": ["setuptools"],
	"tls": ["cryptography", "pyOpenSSL>=17.5.0"],
	"torrent": ["bencode.py>=2.0.0"],  # bencoding ?
	"videofile": ["opencv-python", "av"],
	"win32": ["wmi", "pypiwin32"],
	"widgets": ["wxPython"],
	"xsl": ["lxml"],

	# helpers for specific packages
	"aria": ["aria2p", "requests"],
	"av": ["av"],
	"cv": ["opencv-python", "numpy"],
	"flask": ["flask", "werkzeug>=0.11.11"],
	"gensim": ["gensim>=3.3.0", "numpy"],
	"msgpack": ["msgpack>=0.6.0"],
	"mediainfo": ["pymediainfo"],
	"numpy": ["numpy"],
	"numba": ["numba"],
	"nltk": ["nltk; python_version>='3'", "nltk<3.5; python_version<'3'"],
	"networkx": ["networkx"],
	"pandas": ["pandas"],
	"pdf": ["PyPDF2"],
	"pillow": ["pillow", "piexif"],
	"powerpoint": ["python-pptx"],
	"rasa": ["requests"],
	"salesforce": ["simple_salesforce>=1.1.0", "simplejson"],
	"scipy": ["numpy", "scipy"],
	"toml": ["toml"],
	"torch": ["torch"],
	"twisted": ["twisted"],
	"tensorflow": ["tensorflow"],
	"yaml": ["ruamel.yaml"],
}

extras_require["all"] = sorted(set(chain.from_iterable(extras_require.values())))

setup(
	author="Dobatymo",
	name="genutility",
	version="0.0.42",
	url="https://github.com/Dobatymo/genutility",
	description="A collection of various Python utilities",
	long_description=long_description,
	long_description_content_type="text/markdown",
	classifiers=[
		"Programming Language :: Python :: 2",
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: ISC License (ISCL)",
		"Operating System :: OS Independent",
		"Topic :: Utilities",
	],
	packages=["genutility", "genutility/compat", "genutility/fileformats", "genutility/hardware", "genutility/twothree"],
	package_data={str("genutility"): ["data/*.tsv"]},
	python_requires=">=2.7",
	install_requires=[
		"future",
		"typing;python_version<'3.5'",
		"scandir;python_version<'3.5'",
		"pathlib2;python_version<'3.4'",
		"mock;python_version<'3.3'",
		"contextlib2;python_version<'3.3'",
	],
	extras_require=extras_require,
	use_2to3=False
)



from __future__ import absolute_import, division, print_function, unicode_literals

import os, re, struct, warnings
from io import open
from typing import TYPE_CHECKING

import pkg_resources

from .exceptions import ParseError, Break
from .file import read_or_raise
from .csv import iter_csv
from .compat.os import fspath

if TYPE_CHECKING:
	from typing import BytesIO, Iterator, Tuple

# http://mp4ra.org/#/atoms
# https://wiki.multimedia.cx/index.php/QuickTime_container#QuickTime_Atom_Reference
# https://sno.phy.queensu.ca/~phil/exiftool/TagNames/QuickTime.html

""" `stsd` can be both, be a parent and a leaf atom.
"""

def _load_atoms():
	# type: () -> dict

	package_name = __package__ or "genutility"
	atoms_path = pkg_resources.resource_filename(package_name, "data/mp4-atoms.tsv")

	out = {"container": {}, "leaf": {}}

	for _, type, fourcc, description, _ in iter_csv(atoms_path, delimiter="\t"):
		out[type][fourcc] = description

	return out

atoms = _load_atoms()

atomcodep = re.compile(br"[0-9a-zA-Z]{4}") # what do the specs say here?

def read_atom(fin):
	# type: (BytesIO, ) -> Tuple[int, str, int]

	pos = fin.tell()

	size, code = struct.unpack(">L4s", read_or_raise(fin, 8))

	if not atomcodep.match(code):
		raise ParseError("{!r} @ {} is not a valid atom code".format(code, pos))

	code = code.decode("ascii") # cannot fail

	if size == 1: # 64bit size
		size, = struct.unpack(">Q", read_or_raise(fin, 8))

	return pos, code, size

def _enum_atoms(fin, total_size, depth):
	# type: (BytesIO, int, int) -> Iterator[Tuple[int, int, str, int]]

	while fin.tell() < total_size:
		pos, type, size = read_atom(fin)

		yield depth, pos, type, size

		if size == 0:
			raise Break("Atom extends to the end of the file") # just stop parsing here

		atom_end = pos + size

		if type in atoms["container"]:
			for atom in _enum_atoms(fin, atom_end, depth+1):
				yield atom

		elif type in atoms["leaf"]:
			fin.seek(atom_end, os.SEEK_SET)

		else: # treat it as leaf and skip it
			warnings.warn("Unknown atom: '{}'. Skipping...".format(type), stacklevel=2)
			fin.seek(atom_end, os.SEEK_SET)

	if fin.tell() != total_size:
		raise ParseError("Invalid file structure. Possibly truncated.")

def enumerate_atoms(path):
	# type: (str, ) -> Iterator[Tuple[int, int, str, int]]

	""" Takes an mp4 file `path` and yields (depth, position, code, size) tuples.
		Unknown atoms will print a warning.
	"""

	path = fspath(path)

	total_size = os.path.getsize(path)
	with open(path, "rb") as fr:
		try:
			for atom in _enum_atoms(fr, total_size, 0):
				yield atom
		except Break:
			pass
		except EOFError:
			raise ParseError("Truncated file.")

if __name__ == "__main__":

	import logging
	from argparse import ArgumentParser
	from genutility.args import is_dir
	from genutility.iter import list_except

	parser = ArgumentParser()
	parser.add_argument("path", type=is_dir)
	parser.add_argument("-e", "--errors-only", action="store_true")
	parser.add_argument("-r", "--recursive", action="store_true")
	args = parser.parse_args()

	if args.recursive:
		paths = args.path.rglob("*.mp4")
	else:
		paths = args.path.glob("*.mp4")

	for path in paths:
		if args.errors_only:
			exc, res = list_except(enumerate_atoms(path))
			if exc:
				for depth, pos, type, size in res:
					print("--"*depth, pos, type, size)
				logging.exception("Enumerating atoms of %s failed", path, exc_info=exc)
		else:
			print(path)
			for depth, pos, type, size in enumerate_atoms(path):
				print("--"*depth, pos, type, size)
			print()



from __future__ import absolute_import, division, print_function, unicode_literals

import socket, re
from typing import TYPE_CHECKING

import netifaces

if TYPE_CHECKING:
	from typing import Optional

def is_ipv4(s):
	# type: (str, ) -> bool

	""" Tests if `s` is a IPv4 address string.
		It only validates the common `x.x.x.x` format and rejects less common ones like `x.x`.
	"""

	try:
		nums = tuple(map(int, s.split(".")))
		return len(nums) == 4 and min(nums) >= 0 and max(nums) <= 255
	except (AttributeError, ValueError):
		return False

simple_email_regex = r"[^@\s]+@[^@\s]+\.[^@\s]+"

def is_email(s):
	# type: (str, ) -> bool

	""" Coarse check to test of a string is an email. Will accept some invalid ones like
		"asd@.asd.com" and reject some valid ones like "asd@localhost"
	"""

	return re.fullmatch(simple_email_regex, s)

def get_standard_gateway(default=None):
	# type: (Optional[str], ) -> Optional[str]

	""" Returns the standard IPv4 gateway.
	"""

	for ipv4_gateway in netifaces.gateways()["default"][netifaces.AF_INET]:
		try:
			socket.inet_aton(ipv4_gateway)
			return ipv4_gateway
		except socket.error:
			pass

	return default



from typing import TYPE_CHECKING

from werkzeug.exceptions import BadRequest
from werkzeug.wsgi import get_host

if TYPE_CHECKING:
	from typing import Any, Dict, Optional, Callable, Mapping, Iterable
	WsgiApp = Any

class HostDispatcher(object):

	""" Dispatch WSGI apps by host.
		app = HostDispatcher({'a.com': app_a, 'b.com': app_b})
	"""

	def __init__(self, hosts, default=None):
		# type: (Dict[str, WsgiApp], Optional[WsgiApp]) -> None

		self.hosts = hosts
		self.default = default or BadRequest() # rfc7230#section-5.4

	def __call__(self, environ, start_response):
		# type: (Mapping, Callable) -> Iterable[bytes]

		host = get_host(environ)
		app = self.hosts.get(host, self.default)
		return app(environ, start_response)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.moves.itertools import zip_longest
from future.utils import raise_from
from builtins import zip, map, range, reversed

import sys, logging, random
from collections import deque
from itertools import groupby, chain, tee, islice, count, combinations, product, starmap, repeat
from operator import add
from time import time, sleep
from random import randrange
from types import GeneratorType
from typing import TYPE_CHECKING

from .exceptions import IteratorExhausted, EmptyIterable
from .ops import operator_in

if TYPE_CHECKING:
	from typing import (Any, Callable, Iterable, Iterator, TypeVar, Tuple, Sequence,
		Optional, Collection, List, Union, Type, Dict, TextIO)
	from logging import Logger
	from .typing import SizedIterable, Number
	T = TypeVar("T")
	U = TypeVar("U")
	V = TypeVar("V")

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

def iterrandrange(a, b):
	# type: (int, int) -> Iterator[int]

	""" Yields a stream of non-cryptographic random numbers between `a` (inclusive) and `b` (exclusive).
	"""

	while True:
		yield randrange(a, b)  # nosec

def repeatfunc(func, times=None, *args):
	# type: (Callable[[*Any], T], Optional[int], *any) -> Iterator[T]

	""" Repeat calls to func with specified arguments.
		Example:  repeatfunc(random.random)
		See: https://docs.python.org/3/library/itertools.html#itertools-recipes
	"""

	if times is None:
		return starmap(func, repeat(args))
	return starmap(func, repeat(args, times))

def _lstr(it, length=None):
	if length is None:
		try:
			length = len(it) # type: ignore # TypeError caught below
		except TypeError:
			pass

	if length is not None:
		lstr = "/{}".format(length)
	else:
		lstr = ""

	return length, lstr

def progressdata(it, length=None, refresh=1, end="\r", file=sys.stdout):
	# type: (Union[Iterable, Sequence], Optional[int], float, str, Optional[TextIO]) -> Iterable

	length, lstr = _lstr(it, length)
	last = start = time()
	total = 0

	try:
		for elm in it:
			yield elm
			total += len(elm)
			current = time()
			if current - last > refresh:
				last = current
				duration = current-start
				print("Read {}{}, running for {} seconds ({:0.2e}/s).".format(total, lstr, int(duration), total/duration), end="\r", file=file)
	except KeyboardInterrupt:
		print("Unsafely aborted after reading {}{} in {} seconds ({:0.2e}/s).".format(total, lstr, int(last-start), total/(last-start)), end=end, file=file)
		raise
	else:
		duration = last-start
		if duration > 0:
			print("Finished {} in {} seconds ({:0.2e}/s).".format(total, int(duration), total/duration), end=end, file=file)
		else:
			print("Finished {} in {} seconds.".format(total, int(duration)), end=end, file=file)

def progress(it, length=None, refresh=1, end="\r", file=sys.stdout, extra_info_callback=None):
	# type: (Union[Iterable, Collection], Optional[int], float, str, Optional[TextIO], Callable) -> Iterator

	""" Wraps an iterable `it` to periodically print the progress every `refresh` seconds. """

	""" todo: `from operator import length_hint`
		Use `length_hint(it)` for a guess which might be better than nothing.
		Can return 0.
	"""

	length, lstr = _lstr(it, length)
	extra = ""
	last = start = time()
	i = 0

	for i, elm in enumerate(it, 1):
		yield elm
		current = time()
		if current - last > refresh:
			last = current
			duration = current-start
			if extra_info_callback:
				extra = " [{}]".format(extra_info_callback(i, length))
			print("{}{}, running for {} seconds ({:0.2e}/s){}.".format(i, lstr, int(duration), i/duration, extra), end="\r", file=file)

	duration = last-start
	if duration > 0:
		print("Finished {} in {} seconds ({:0.2e}/s).".format(i, int(duration), i/duration), end=end, file=file)
	else:
		print("Finished {} in {} seconds.".format(i, int(duration)), end=end, file=file)

class Progress(object):

	""" Cannot create a decorator to do this to generic generators,
		because attributes cannot be set on generator objects.
	"""

	def __init__(self, obj):
		self.obj = obj

	def __iter__(self):
		return progress(self.obj)

	def __len__(self):
		# type: () -> int

		""" Can raise a TypeError if underlying object does not support `len()`.
		"""

		return len(self.obj)

def count_distinct(it):
	# type: (Iterable[Any], ) -> int

	return len(set(it))

def reversedzip(*its):
	return zip(*map(reversed, its))

def flatten(it):
	# type: (Union[tuple, list], ) -> Iterator

	for elm in it:
		if isinstance(elm, (tuple, list)):
			for i in flatten(elm):
				yield i
		else:
			yield elm

def diffsgn(a, b):
	# type: (Number, Number) -> int

	""" Sign of the finite derivative.
		Equal to `-cmp(a, b)` or `cmp(b, a).
	"""

	return (a < b) - (a > b)

def asc_peaks(iterable):
	# type: (Iterable[T], ) -> Iterator[T]

	""" Yields ascending peaks from `iterable`. """

	it = iter(iterable)
	try:
		cur = next(it)
	except StopIteration:
		return

	for x in it:
		if x < cur:
			yield cur
		cur = x
	yield cur

def extrema(iterable, first={1, -1}, last={1, -1}, derivatives={1, -1}):
	# type: (Iterable[T], Set[int], Set[int], Set[int]) -> Iterator[T]

	it = iter(iterable)
	try:
		old = next(it)
	except StopIteration:
		return
	try:
		old_2 = next(it)
		old_d = diffsgn(old, old_2)
	except StopIteration:
		yield old # fixme: should this be returned?
		return

	if old_d in first:
		yield old

	old = old_2

	for new in it:
		new_d = diffsgn(old, new)
		if new_d != old_d and old_d in derivatives:
			yield old
		old = new
		old_d = new_d

	if old_d in last:
		yield old

def peaks(it):
	# type: (Iterable[T], ) -> Iterator[T]

	""" Yields peaks of `it`. """

	return extrema(it, {-1}, {1}, {1})

def valleys(it):
	# type: (Iterable[T], ) -> Iterator[T]

	""" Yields valleys of `it`. """

	return extrema(it, {1}, {-1}, {-1})

def empty():
	# type: () -> Iterator

	""" An empty iterator. Other methods include:
		iter(())
		yield from ()
	"""

	return
	yield # pylint: disable=unreachable

def lastdefault(it, default=None):
	# type: (Iterable[T], Optional[T]) -> Optional[T]

	""" Returns the last element of iterable `it`. It discards all other values.
		This method is about 1/2 times as fast as `last()` for large iterables.
	"""

	ret = default
	for i in it:
		ret = i
	return ret

def last(it):
	# type: (Iterable[T], ) -> T

	""" Returns the last element of iterable `it`. It discards all other values.
		This method is about 2 times as fast as `lastdefault()` for large iterables.
	"""

	try:
		return deque(it, 1).pop()
	except IndexError:
		raise_from(EmptyIterable("Empty iterable"), None)

def batch(it, n, func=None):
	# type: (Iterable[Any], int, Optional[Callable]) -> Iterator[Any] # return type cannot be more specific because of filter()

	""" Batches iterable `it` into batches of size `n`.
		Optionally post-processes batch with `func`.
	"""

	it = iter(it)
	while True:
		chunk_it = islice(it, n)
		try:
			first_el = next(chunk_it)
		except StopIteration:
			return

		batch_iter = chain((first_el,), chunk_it)

		if func:
			batch_iter = func(batch_iter)

		yield batch_iter

def advance(it, n):
	# type: (Iterable[T], int) -> None

	""" Advances the iterable `it` n steps. """

	"""
	for i in range(n):
		next(it, None)
	"""
	next(islice(it, n, n), None)

def filternone(it):
	# type: (Iterable, ) -> Iterator

	""" Removes all `None` values from `it`. """

	return (i for i in it if i is not None)

def all_equal(it):
	# type: (Iterable, ) -> bool

	""" Returns `True` if all elements of `it` are equal to each other.

		Much faster than:
			return len(set(it)) in (0, 1)
		and also faster than:
			try:
				first = next(it)
			except StopIteration:
				return True
			return all(first == rest for rest in it)
		for both extreme cases: all equal and none equal
	"""

	g = groupby(it)
	return next(g, True) and not next(g, False) # type: ignore

def pairwise(it):
	# type: (Iterable[T], ) -> Iterator[Tuple[T, T]]

	""" Return two elements of iterable `it` at a time,
		but only advances the iterable by 1 each time.
		pairwise((a, b, c)) -> (a, b), (b, c)
	"""

	a, b = tee(it, 2)
	next(b, None)
	return zip(a, b)

def findfirst(func, it, default=(None, None)):
	# type: (Callable[[T], bool], Iterable[T], Tuple[Optional[int], Optional[T]]) -> Tuple[Optional[int], Optional[T]]

	""" Find the first element of iterable `it` where `func(elm)` evaluates to True.
		Return `default` if not such element was found.
	"""

	for i, x in enumerate(it):
		if func(x):
			return i, x

	return default

def is_empty(it):
	# type: (Iterator, ) -> bool

	""" Returns True if the iterator `it` is already fully consumed.
		If not, it will be advanced.
	"""

	try:
		next(it)
		return False
	except StopIteration:
		return True

def consume(it):
	# type: (Iterable, ) -> None

	""" Consumes the iterable `it` completely. """

	deque(it, maxlen=0)

def resizer(it, size, pad=False, pad_elm=None):
	# type: (Iterable[Sequence[T]], int, bool, Optional[T]) -> Iterable

	""" Cuts the input iterable `it` consisting of variable length slices
		into slices of length `size`. If pad is True, the last slice
		will be padded to length `size` with `pad_elm`.

		resizer(("asd", "qw", "e"), 4) -> (['a', 's', 'd', 'q'], ['w', 'e'])
		resizer(("asd", "qw", "e"), 4, True, "x") -> (['a', 's', 'd', 'q'], ['w', 'e', 'x', 'x'])
	"""

	if size <= 0:
		raise ValueError("size must be larger than 0")

	it = iter(it)
	buf = "" # type: Sequence[T]
	pos = 0
	try:
		while True:
			to_read = size
			out = []
			while to_read > 0:
				if len(buf) == pos:
					buf = next(it)
					pos = 0
				tmp = buf[pos:pos+to_read]
				out.extend(tmp)
				to_read -= len(tmp)
				pos += len(tmp)
			yield out
	except StopIteration:
		if out:
			while pos < len(buf):
				logger.error("this can never happen, can it?")
				tmp = buf[pos:pos+to_read]
				out.extend(tmp)
				to_read -= len(tmp)
				pos += to_read
			if pad:
				yield out + [pad_elm]*(size-len(out))
			else:
				yield out

def switch(it):
	# type: (Iterable[Tuple[T, U]], ) -> Iterator[Tuple[U, T]]

	""" Swaps the elements in iterable of pairs `it`.
		((3, 6), (7, 4), (1, 9)) -> (6, 3), (4, 7), (9, 1)
	"""

	return ((j, i) for i, j in it)

def switched_enumerate(it):
	# type: (Iterable[T], ) -> Iterator[Tuple[T, int]]

	""" Same as `enumerate()` except that order of the output pair is switched.
	"""

	return switch(enumerate(it))

def multi_join(it_a, it_b, join_func=add):
	# type: (Iterable[T], Iterable[U], Callable[[T, U], V]) -> Iterator[V]

	""" Calls `join_func` on the cross product of `it_a` and `it_b`.
		multi_join((1, 2), (3, 4), lambda x, y: str(x)+str(y)) -> ('13', '23', '14', '24')
	"""

	return [join_func(a, b) for b in it_b for a in it_a]

def iter_except(iterator, exception_callbacks, return_on_exception=False):
	# type: (Iterator[T], Dict[Type, Callable[[Iterator[T], Exception], bool]], bool) -> Iterator[T]

	""" Calls callbacks when exceptions are raised in `iterator`. Does not work for Generators.
		The type key in the `exception_callbacks` dict must be the same type as the exception which
		should be caught, not a subtype.
	"""

	if not return_on_exception and isinstance(iterator, GeneratorType):
		raise TypeError("iterator cannot be a generator")

	while True:
		try:
			yield next(iterator)
		except StopIteration:
			return
		except Exception as e:
			try:
				if exception_callbacks[type(e)](iterator, e): # use if x. and not if not x, so None doesn't raise, if user forgets to returns correct value
					raise
			except KeyError:
				raise e
			if return_on_exception:
				return

def list_except(it, catch=Exception):
	# type: (Iterable[T], Union[Exception, Sequence[Exception]]) -> Tuple[Exception, List[T]]

	""" Same as `list()` except in the case of an exception, the partial list which was collected
		so far is returned. `catch` specifies which exceptions are caught. It can be an exception
		or a tuple of exceptions.
	"""

	ret = [] # List[T]
	exc = None # type: Optional[Exception]

	try:
		for i in it:
			ret.append(i)
	except catch as e:
		exc = e

	return exc, ret

def iter_except_ignore(iterator, exceptions, logger=logger):
	# type: (Iterator[T], Collection[Exception], Logger) -> Iterator[T]

	""" Ignores `exceptions` raised in `iterator`. Does not work for Generators.
	"""

	if isinstance(iterator, GeneratorType):
		raise TypeError("iterator cannot be a generator")

	while True:
		try:
			yield next(iterator)
		except StopIteration:
			break
		except exceptions as e:
			logger.warning("Error in iterable: %s", str(e))

def decompress(selectors, data, default=None):
	# type: (Iterable[bool], Iterator[T], Optional[T]) -> Iterator[T]

	""" Basically the opposite of `itertools.compress`.
		decompress([True, False, True], iter(["A", "B"])) -> ("A", None, "B")
	"""

	try:
		for s in selectors:
			if s:
				yield next(data)
			else:
				yield default
	except StopIteration: # exception converted because StopIterations raised in generators will cause RuntimeErrors
		raise IteratorExhausted

def first_not_none(it, default=None):
	# type: (Iterable[T], Optional[T]) -> Optional[T]

	""" Returns the first element of `it` which is not None. """

	return next(filternone(it), default)

def range_count(start=0, stop=None, step=1):
	# type: (int, Optional[int], int) -> Iterator[int]

	""" Similar to `range`, except it optionally allows for an infinite range. """

	if stop:
		return range(start, stop, step)
	else:
		return count(start, step)

def product_range_repeat(depth, args):
	# type: (int, Tuple) -> Iterator

	""" Returns coordinate or coordinate slices useful for iteration over space.
		Can replace higher dimensional for-in-range loops.
	"""

	return product(tuple(range(*args)), repeat=depth)

def powerset(sit):
	# type: (SizedIterable[T], ) -> Iterator[Tuple[T, ...]]

	""" Returns the powerset of a sized iterable.
		The powerset of an empty set is a set including the empty set.

		powerset([1,2,3]) --> (), (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
	"""

	return chain.from_iterable(combinations(sit, r) for r in range(len(sit)+1))

# had cmp=b"\0" default before
def all_equal_to(it, cmp):
	# type: (Iterable[T], T) -> bool

	""" Test if every item of iterable `it` is equal to `cmp`.
	"""

	return all(elm == cmp for elm in it)

def iter_equal(*its):
	# type: (*Iterable, ) -> bool

	""" Test if the contents of all iterables `its` is the same.
	"""

	return all(all_equal(it_elements) for it_elements in zip_longest(*its))

def iter_different(it_a, it_b):
	# type: (Iterable, Iterable) -> bool

	""" Test if `it_a` and `it_b` yield different elements. """

	return any(a != b for a, b in zip_longest(it_a, it_b))

def every_n(it, n, pos=0):
	# type: (Iterable[T], int, int) -> T

	""" Yields every `n`-th element from iterable `it` and starts at `pos`. """

	if pos > 0:
		advance(it, pos)

	for i, val in enumerate(it):
		if i % n == 0:
			yield val

def split(it, size):
	# type: (Iterable[T], int) -> Sequence[Iterable[T]]

	""" memory consumption should be proportional to `size` when output iterators
		are used in parallel loops (like `zip`), else might be higher
	"""

	copies = tee(it, size)
	return tuple(every_n(it, size, pos) for it, pos in zip(copies, range(size)))

def no_dupes(*its):
	# type: (*Iterable[T], ) -> Iterator[T]

	""" Merge `its` and removes all duplicates while preserving order. """

	seen = set()  # type: Set[T]

	for it in its:
		for x in it:
			if x not in seen:
				yield x
				seen.add(x)

def retrier(waittime, attempts=-1, multiplier=1, jitter=0, max_wait=None, jitter_dist="uniform", waitfunc=sleep):
	# type: (float, int, float, float, Optional[float], str, Callable[[float], Any]) -> Iterator[None]

	""" Iterator which yields after predefined amounts of time.
		Supports `multiplier` to implements linear or exponential backoff.
		`jitter` adds noise to the average wait time, distributed according to `jitter_dist`.
		Supported values for `jitter_dist` are `uniform` and `normal`.
	"""

	if jitter_dist == "uniform":
		rand = random.uniform
	elif jitter_dist == "normal":
		rand = random.normalvariate
	else:
		raise ValueError("Unsupported jitter_dist")

	if attempts == 0:
		return

	for i in count():
		yield i

		attempts -= 1

		if attempts != 0:
			if max_wait and waittime > max_wait:
				jitter *= max_wait/waittime
				waittime = max_wait
				multiplier = 1

			do_wait = max(0, waittime + rand(-jitter, jitter))
			waitfunc(do_wait)
		else:
			break

		waittime *= multiplier
		jitter *= multiplier

def collapse_any(it, col_set):
	# type: (Iterable[T], Set[T]) -> Iterator[T]

	""" Removes consecutive elements from iterable `it` if element is in `col_set`.
		([1,1,2,2,3,3,4,4], {1,2}) -> 1, 2, 3, 3, 4, 4
	"""

	assert isinstance(col_set, set)

	for key, g in groupby(it):
		if key in col_set:
			yield key
		else:
			for i in g: # yield from g
				yield i

def collapse_all(it, col_set, replacement):
	# type: (Iterable[T], Set[T], T) -> Iterator[T]

	""" Removes consecutive elements from iterable `it` if element is in `col_set`
		and replace `replacement`.
		([1,1,2,2,3,3,4,4], {1,2}, 5) -> 5, 3, 3, 4, 4
	"""

	assert isinstance(col_set, set)

	for key, g in groupby(it, key=operator_in(col_set)):
		if key:
			yield replacement
		else:
			for i in g: # yield from g
				yield i

def any_in_common(a, b):
	# type: (Iterable[T], Iterable[T]) -> bool

	""" Tests if iterables `a` and `b` have any elements in common.
		See also: `any_in`.
	"""

	return any(i == j for i, j in product(a, b))

def any_in(it, container):
	# type: (Iterable[T], Container[T]) -> bool

	""" Tests if any elements of `it` are in `container`.
		For best performance, `container` should be a set.
		See also: `any_in_common`.
	"""

	if isinstance(it, set) and isinstance(container, set):
		return not it.isdisjoint(set)

	return any(elm in container for elm in it)

class CountIter(object):

	""" Example:
		c = CountIter()
		with open("asd.txt", "rt", encoding="utf-8") as fr:
			for line in c.count(fr):
				func(line)  # do something with line
		print("Processed", c.get(), "lines")
	"""

	def __init__(self):
		self.i = 0

	def count(self, it):
		for elm in it:
			self.i += 1
			yield elm

	def get(self):
		return self.i

def select_by_indices(it, indices):
	# type: (Iterable[T], Iterator[int]) -> Iterator[T]

	try:
		sel_ind = next(indices)
		for i, elem in enumerate(it):
			if i == sel_ind:
				yield elem
				sel_ind = next(indices)
	except StopIteration:
		pass

def skip_by_indices(it, indices):
	# type: (Iterator[T], Iterator[int]) -> Iterator[int, T]

	i = 0
	try:
		sel_ind = next(indices)
		for elem in it:
			if i == sel_ind:
				sel_ind = next(indices)
				continue

			yield i, elem
			i += 1
	except StopIteration:
		pass

	for ii, elem in enumerate(it, i):
		yield ii, elem



from __future__ import absolute_import, division, print_function, unicode_literals

try:
	from builtins import FileNotFoundError
except ImportError:
	FileNotFoundError = IOError

try:
	from builtins import FileExistsError
except ImportError:
	FileExistsError = IOError



from __future__ import absolute_import, division, print_function, unicode_literals

from hashlib import sha1
from typing import TYPE_CHECKING

from bencode import bdecode, bencode

from .file import read_file, blockfileiter, blockfilesiter
from .filesystem import FileProperties
from .exceptions import assert_choice

if TYPE_CHECKING:
	from typing import Optional
	from .compat.pathlib import Path

def read_torrent(path):
	return bdecode(read_file(path, "rb"))

def read_torrent_info_dict(path):
	return bdecode(read_file(path, "rb"))["info"]

def iter_torrent(path):
	""" Returns file informations from a torrent file. 
		Hash (SHA-1) is obtained according to BEP0047 (if available).
	"""

	info = read_torrent_info_dict(path)

	try:
		files = info["files"]
	except KeyError:
		yield FileProperties(info["name"], info["length"], False, hash=info.get("sha1"))
	else:
		for fd in files:
			path = "/".join(fd["path"])
			yield FileProperties(info["name"] + "/" + path, fd["length"], False, hash=fd.get("sha1"))

def pieces_field(pieces):
	return b"".join(sha1(piece).digest() for piece in pieces)  # nosec

def create_torrent_info_dict(path, piece_size, private=None):
	# type: (Path, int, Optional[int]) -> dict

	if private is not None:
		assert_choice("private", private, (0, 1))

	if path.is_file():
		ret = {
			"name": path.name,
			"length": path.stat().st_size,
			"piece length": piece_size,
			"pieces": pieces_field(blockfileiter(path, chunk_size=piece_size)),
		}

	elif path.is_dir():
		files = []

		assert files

		ret = {
			"name": path.name,
			"files": [{
				"length": length,
				"path": path.parts,
			} for length, path in files],
			"piece length": piece_size,
			"pieces": pieces_field(blockfilesiter((p for l, p in files), chunk_size=piece_size)),
		}

	else:
		raise ValueError("path neither file nor directory")

	if private is not None:
		ret["private"] = private

	return ret

def torrent_info_hash(d):
	return sha1(bencode(d)).hexdigest()  # nosec

def get_torrent_hash(path):
	return torrent_info_hash(read_torrent_info_dict(path))

def create_torrent(path, piece_size, announce=""):
	# type: (Path, int, str) -> bytes

	info = create_torrent_info_dict(path, piece_size)
	torrent = {
		"info": info,
		"announce": announce
	}

	return bencode(torrent)



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import List
	from .typing import Number

class VariableRowMatrix(object):

	def __init__(self):
		# type: () -> None

		self.lol = []

	@classmethod
	def from_list_of_lists(cls, lol):
		# type: (List[List[Number]], ) -> VariableRowMatrix

		m = VariableRowMatrix.__new__(cls)
		m.lol = lol
		return m

	def __len__(self):
		# type: () -> int

		return len(self.lol)

	def __setitem__(self, key, value):
		i, j = key

		try:
			row = self.lol[i]
		except IndexError:
			mul = i - len(self.lol) + 1
			self.lol.extend(mul*[[]])
			row = self.lol[i]

		try:
			row[j] = value
		except IndexError:
			mul = j - len(row) + 1
			row.extend(mul*[0])
			row[j] = value

	def __getitem__(self, key):
		i, j = key
		return self.lol[i][j]



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import map, range
from future.utils import viewvalues, raise_from

from functools import partial
from math import pi, factorial, sqrt, log
from collections import defaultdict, Counter
from itertools import islice
from operator import lt, gt, mul
from typing import TYPE_CHECKING
from random import sample

from .compat.math import prod, isqrt, nan as NaN
from .iter import range_count
from .exceptions import EmptyIterable

if TYPE_CHECKING:
	from typing import TypeVar, Iterable, Tuple, Optional, Callable, Sequence, SupportsFloat
	from .typing import Computable, Orderable
	O = TypeVar("O", bound=Orderable)

inf = float("inf")

def reciprocal(n):
	# type: (Computable, ) -> Computable

	return 1. / n

def sqr(x):
	# type: (Computable, ) -> Computable

	""" Returns `x` to the power of 2.
	"""

	return x*x

def dot(a, b):
	# type: (Iterable[Computable], Iterable[Computable]) -> Computable

	""" Returns the dot product (inner product) of vectors `a` and `b`.
	"""

	return sum(map(mul, a, b))

# was: euclidic_norm
def euclidean_norm(x):
	# type: (Iterable[Computable], ) -> Computable

	""" Returns the Euclidean norm of vector `x`.
	"""

	return sqrt(sum(map(sqr, x)))

def shannon_entropy(ps, base=2):
	# type: (Iterable[float], int) -> float

	""" Calculates the Shannon entropy for probabilities `ps` with `base` in pure Python. """

	return -sum(p * log(p, base) for p in ps)

def cosine_similarity(a, b):
	# type: (Iterable[Computable], Iterable[Computable]) -> Computable

	""" Calculate the cosine similarity (normalized dot product) of vectors `a` and `b`.
	"""

	return dot(a, b)/euclidean_norm(a)/euclidean_norm(b)

def argmin(l, s=0, e=None):
	# type: (Sequence[O], int, Optional[int]) -> O

	if not l:
		raise EmptyIterable()

	arg = s
	e = e or len(l)
	for i in range(s+1, e):
		if l[i] < l[arg]:
			arg = i
	return arg

def argmax(l, s=0, e=None):
	# type: (Sequence[O], int, Optional[int]) -> O

	if not l:
		raise EmptyIterable()

	arg = s
	e = e or len(l)
	for i in range(s+1, e):
		if l[i] > l[arg]:
			arg = i
	return arg

def argmax_pair(iterable):
	# type: (Iterable[O], ) -> Tuple[int, O]

	it = iter(iterable)
	arg = 0
	try:
		max = next(it)
	except StopIteration:
		raise_from(EmptyIterable("empty iterable"), None)

	for i, elm in enumerate(it, 1):
		if elm > max:
			arg = i
			max = elm

	return arg, max

from itertools import count
from operator import itemgetter
def argmax_v2(iterable):
	# type: (Iterable[O], ) -> Tuple[int, O]

	""" nicer, but almost 2 times slower than above"""
	return max(zip(count(), iterable), key=itemgetter(1))[0]

def minmax(a, b):
	# type: (O, O) -> O

	""" `default` argument cannot be used, because the C level function doesn't have a default value
		but an overload.
	"""

	if a <= b:
		return a, b
	else:
		return b, a

def _argfind_cmp(it, target, op1, op2, pos=0):
	# type: (Iterable[O], O, Callable[[O, O], bool], Callable[[O, O], bool], int) -> Tuple[int, O]

	""" Find the largest element in `it` less than or equal to `target` and return the index and value.
		Return (-1, None) if no such element can be found.
	"""

	idx = -1
	val = None

	it = islice(it, pos, None)
	try:
		v = next(it)
	except StopIteration:
		return idx, val

	if v == target:
		return pos, v
	elif op1(v, target):
		idx = pos
		val = v

	for i, v in enumerate(it, pos+1):
		if v == target:
			return i, v
		elif op1(v, target) and (val is None or op2(v, val)):
			idx = i
			val = v

	return idx, val

def argfind_lte(it, target, pos=0):
	return _argfind_cmp(it, target, lt, gt, pos)

def argfind_gte(it, target, pos=0):
	return _argfind_cmp(it, target, gt, lt, pos)

def degree_to_rad(angle):
	# type: (float, ) -> float

	return 2.*pi*angle/360.

def limit(x, left=0, right=1):
	if x >= left:
		if x <= right:
			return x
		else:
			return right
	else:
		return left

def relative_luminance(r, g, b):
	# type: (float, float, float) -> float

	""" Converts linear RGB components to relative luminance.
		See: https://en.wikipedia.org/wiki/Relative_luminance
	"""

	return 0.2126*r + 0.7152*g + 0.0722*b

# sys.maxsize
class _PosInfInt():

	def __lt__(self, rhs):
		return False

	def __gt__(self, rhs):
		return True

	def __eq__(self, rhs):
		raise ArithmeticError("don't compare to other PosInfInt...")

	def __add__(self, rhs):
		return self

	def __sub__(self, rhs):
		return self

	def __str__(self):
		return "PosInfInt"

	def __repr__(self):
		return "PosInfInt"

PosInfInt = _PosInfInt()

def multinomial_coefficient(n, ks):
	# type: (int, Iterable[int]) -> int

	return factorial(n) / prod(map(factorial, ks))

def num_unique_permutations(word):
	# type: (str, ) -> int

	""" Number of unique permutations of `word`. """

	return multinomial_coefficient(len(word), viewvalues(Counter(word)))

def discrete_distribution(it):
	# type: (Iterable, ) -> Tuple[defaultdict, int]

	""" Creates a discrete distribution of the elements of `it`. """

	d = defaultdict(int)
	s = 0
	for i in it:
		s += 1
		d[i] += 1
	return (d, s)

def fibonaccigen(f0=0, f1=1):
	# type: (int, int) -> int
	""" A000045 [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, ...]
		arguments are the first two element of the Fibonacci series

		To get a specific Fibonacci number use `fibonacci(n)`.
	"""

	yield f0
	while True:
		yield f1
		f1, f0 = f1+f0, f1

def floordiv2(n):
	while n > 0:
		yield n
		n = n // 2

def isintlog(num, base):
	# type: (int, int) -> bool

	if num <= 0 or base <= 0:
		raise ValueError("num and base must be larger than 0")

	if num == 1:
		return False

	while num > 1:
		r = num % base
		if r != 0:
			return False
		num //= base

	return True

def _fibonacci(n):
	# returns (F(n), F(n+1)).
	# inspired by https://www.nayuki.io/page/fast-fibonacci-algorithms

	a, b = 0, 1
	for i in reversed(tuple(floordiv2(n))):
		c = a * (b * 2 - a)
		d = a * a + b * b
		if i % 2 == 0:
			a = c
			b = d
		else:
			a = d
			b = c + d
	return a, b

#from .numba import opjit
#@opjit() # doesn't support arbitrary large ints
def _fibonacci_v2(n):
	#see: https://stackoverflow.com/a/1526036
	# 'Structure and Interpretation of Computer Programs' (SICP) ex 1.19
	# http://community.schemewiki.org/?sicp-ex-1.19

	a, b, p, q = 1, 0, 0, 1
	while n > 0:
		if n % 2 == 0: # even
			oldp = p
			p = p*p + q*q
			q = 2*oldp*q + q*q
			n //= 2
		else:
			olda = a
			a = b*q + a*q + a*p
			b = b*p + olda*q
			n -= 1
	return a, b

def fibonacci(n):

	""" implementation of "fast doubling". can be derived from this identity:
		[[1, 1], [1, 0]]^n = [[F_n+1, F_n], [F_n, F_n-1]]
	"""

	if n == 0:
		return 0
	elif n < 0:
		raise ValueError("Negative arguments not implemented")

	return _fibonacci(n-1)[1]

def byte2size(byte, exp=0, base=1024):
	# type: (SupportsFloat, int, int) -> Tuple[float, str]

	""" Converts integer number to float and unit string """

	byte = float(byte)
	byte_units = ("Byte", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")

	for unit in islice(byte_units, exp, None):
		if (byte >= base):
			byte /= base
		else:
			return byte, unit

def byte2size_str(byte, roundval=3):
	# type: (SupportsFloat, int) -> str

	""" Converts integer number to human readable size string. """

	byte, value = byte2size(byte)
	if value == "Byte":
		roundval = 0
	return "{:.{}f} {}".format(byte, roundval, value)

def primes(stop=None):
	# type: (Optional[int], ) -> Iterable[int]

	""" Naive algorithm to yield all primes which are smaller than `stop`.
		If `stop` is None it will never stop generating primes.
		For better algorithms see here: https://stackoverflow.com/questions/2211990/how-to-implement-an-efficient-infinite-generator-of-prime-numbers-in-python/10733621#10733621
	"""

	found_primes = []
	first_prime = 2

	for i in range_count(first_prime, stop):
		is_prime = True
		for j in found_primes:
			if j > isqrt(i):
				break
			if i % j == 0:
				is_prime = False
				break

		if is_prime:
			found_primes.append(i)
			yield i

def additaet(n):
	# type: (int, ) -> int

	""" sums all numbers from 0 to n using Gaussian formula.
	basically same performance
		(n*n + n) >> 1
	too slow, floats
		int(((n/2.0) + 0.5) * n)
		int((n + 1) / 2.0 * n)
	too slow, worse complexity
		sum(range(n+1))
	"""

	return ((n + 1) * n) >> 1

addity = additaet

def digitsum(n):
	# type: (int, ) -> int

	""" Use for numbers with more than 6 digits. """

	return sum(map(int, str(n)))

def digitsum_small(n):
	# type: (int, ) -> int

	""" Use for numbers with less than 6 digits. """

	s = 0
	while n:
		s += n % 10
		n //= 10
	return s

def digitsum_base(n, base=10):
	# type: (int, int) -> int

	""" Use with string input if base is not 10. """

	return sum(map(partial(int, base=base), str(n)))

def euclidean_distance(a, b):
	# type: (Computable, Computable) -> Computable

	""" Euclidean distance between `a` and `b`. """

	return abs(b-a)

def closest(numbers, number, distance_func=euclidean_distance):
	# type: (Iterable, Computable) -> Computable

	""" For a list of `numbers`, return the closest number to `number`. """

	return min(numbers, key=partial(distance_func, number))

def absolute_difference(a, b):
	# type: (Computable, Computable) -> Computable

	return abs(b - a)

def sortedsample(n, k):
	# type: (int, int) -> List[int]

	""" Sorted random sample without replacements. """

	return sorted(sample(range(n), k))

number_metric = absolute_difference

if __name__ == "__main__":
	import timeit
	print(min(timeit.repeat(stmt="digitsum(111111)", setup="from __main__ import digitsum")))
	print(min(timeit.repeat(stmt="digitsum_base(111111)", setup="from __main__ import digitsum_base")))
	#print(min(timeit.repeat('_fibonacci_v2(999999999999999999999999999999999)', setup="from __main__ import _fibonacci_v2")))
